"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.sizeSnapshot = undefined;

var _fs = require("fs");

var _path = require("path");

var _rollup = require("rollup");

var _uglifyEs = require("uglify-es");

var _gzipSize = require("gzip-size");

var _gzipSize2 = _interopRequireDefault(_gzipSize);

var _fastDeepEqual = require("fast-deep-equal");

var _fastDeepEqual2 = _interopRequireDefault(_fastDeepEqual);

var _jestDiff = require("jest-diff");

var _jestDiff2 = _interopRequireDefault(_jestDiff);

var _bytes = require("bytes");

var _bytes2 = _interopRequireDefault(_bytes);

var _chalk = require("chalk");

var _chalk2 = _interopRequireDefault(_chalk);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const validateOptions = options => {
  const optionsKeys = ["treeshake", "snapshotPath", "updateSnapshot", "printInfo"];

  const invalidKeys = Object.keys(options).filter(d => !optionsKeys.includes(d));

  if (invalidKeys.length) {
    throw Error(`Options ${invalidKeys.map(d => `"${d}"`).join(", ")} are invalid`);
  }
};

const readJsonSync = file => {
  try {
    const text = (0, _fs.readFileSync)(file, "utf-8");
    return JSON.parse(text);
  } catch (error) {
    return {};
  }
};

const treeshakeSize = code => {
  const input = "__size_snapshot_input__";
  const entry = "__size_snapshot_entry__";

  const isReservedId = id => id === input || id === entry;

  const config = {
    input,
    onwarn() {},
    plugins: [{
      resolveId(importee) {
        if (isReservedId(importee)) {
          return importee;
        }
        return null;
      },

      load(id) {
        if (id === input) {
          return `import {} from "${entry}";`;
        }
        if (id === entry) {
          return code;
        }
        return null;
      }
    }]
  };

  return (0, _rollup.rollup)(config).then(bundle => bundle.generate({ format: "es" })).then(result => (0, _uglifyEs.minify)(result.code, { toplevel: true })).then(result => result.code.length);
};

const bytesConfig = { thousandsSeparator: ",", unitSeparator: " ", unit: "B" };

const formatSize = d => _chalk2.default.bold(_bytes2.default.format(d, bytesConfig));

const sizeSnapshot = exports.sizeSnapshot = (options = {}) => {
  validateOptions(options);

  const shouldTreeshake = options.treeshake === true;
  const snapshotPath = options.snapshotPath || (0, _path.join)(process.cwd(), ".size-snapshot.json");
  const shouldUpdateSnapshot = options.updateSnapshot !== false;
  const shouldPrintInfo = options.printInfo !== false;

  return {
    name: "size-snapshot",

    transformBundle(source, outputOptions) {
      const output = outputOptions.file;

      if (typeof output !== "string") {
        throw Error("output file in rollup options should be specified");
      }

      const minified = (0, _uglifyEs.minify)(source).code;

      return Promise.all([(0, _gzipSize2.default)(minified), shouldTreeshake ? treeshakeSize(source) : 0]).then(([gzippedSize, treeshakedSize]) => {
        const sizes = {
          bundled: source.length,
          minified: minified.length,
          gzipped: gzippedSize
        };

        let infoString = `Computed sizes of "${output}"\n` + `  bundled: ${formatSize(sizes.bundled)}\n` + `  minified with uglify: ${formatSize(sizes.minified)}\n` + `  minified and gzipped: ${formatSize(sizes.gzipped)}\n`;

        if (shouldTreeshake) {
          sizes.treeshaked = treeshakedSize;
          const msg = "treeshaked and uglified with toplevel option";
          infoString += `  ${msg}: ${formatSize(treeshakedSize)}\n`;
        }

        const snapshot = readJsonSync(snapshotPath);
        if (shouldUpdateSnapshot) {
          if (shouldPrintInfo) {
            console.info(infoString);
          }
          snapshot[outputOptions.file] = sizes;
          (0, _fs.writeFileSync)(snapshotPath, JSON.stringify(snapshot, null, 2));
        } else {
          const entry = snapshot[output] || {};
          if (!(0, _fastDeepEqual2.default)(entry, sizes)) {
            console.error((0, _jestDiff2.default)(entry, sizes));
            throw Error("size snapshot is not matched");
          }
        }

        return null;
      });
    }
  };
};